@@ -1,6 +1,7 @@
 DEFAULT_JOB_CLASS = 'rq.job.Job'
 DEFAULT_QUEUE_CLASS = 'rq.Queue'
 DEFAULT_WORKER_CLASS = 'rq.Worker'
+DEFAULT_SERIALIZER_CLASS = 'rq.serializers.DefaultSerializer'
 DEFAULT_CONNECTION_CLASS = 'redis.Redis'
 DEFAULT_WORKER_TTL = 420
 DEFAULT_JOB_MONITORING_INTERVAL = 30
@@ -8,3 +9,4 @@ DEFAULT_RESULT_TTL = 500
 DEFAULT_FAILURE_TTL = 31536000  # 1 year in seconds
 DEFAULT_LOGGING_DATE_FORMAT = '%H:%M:%S'
 DEFAULT_LOGGING_FORMAT = '%(asctime)s %(message)s'
+CALLBACK_TIMEOUT = 60
