@@ -1,4 +1,5 @@
 import calendar
+from rq.serializers import resolve_serializer
 import time
 from datetime import datetime, timedelta, timezone
 
@@ -11,7 +12,7 @@ from .queue import Queue
 from .utils import backend_class, current_timestamp
 
 
-class BaseRegistry(object):
+class BaseRegistry:
     """
     Base implementation of a job registry, implemented in Redis sorted set.
     Each job is stored as a key in the registry, scored by expiration time
@@ -21,13 +22,15 @@ class BaseRegistry(object):
     key_template = 'rq:registry:{0}'
 
     def __init__(self, name='default', connection=None, job_class=None,
-                 queue=None):
+                 queue=None, serializer=None):
         if queue:
             self.name = queue.name
             self.connection = resolve_connection(queue.connection)
+            self.serializer = queue.serializer
         else:
             self.name = name
             self.connection = resolve_connection(connection)
+            self.serializer = resolve_serializer(serializer)
 
         self.key = self.key_template.format(self.name)
         self.job_class = backend_class(self, 'job_class', override=job_class)
@@ -58,15 +61,15 @@ class BaseRegistry(object):
         self.cleanup()
         return self.connection.zcard(self.key)
 
-    def add(self, job, ttl=0, pipeline=None):
+    def add(self, job, ttl=0, pipeline=None, xx=False):
         """Adds a job to a registry with expiry time of now + ttl, unless it's -1 which is set to +inf"""
         score = ttl if ttl < 0 else current_timestamp() + ttl
         if score == -1:
             score = '+inf'
         if pipeline is not None:
-            return pipeline.zadd(self.key, {job.id: score})
+            return pipeline.zadd(self.key, {job.id: score}, xx=xx)
 
-        return self.connection.zadd(self.key, {job.id: score})
+        return self.connection.zadd(self.key, {job.id: score}, xx=xx)
 
     def remove(self, job, pipeline=None, delete_job=False):
         """Removes job from registry and deletes it if `delete_job == True`"""
@@ -77,7 +80,7 @@ class BaseRegistry(object):
             if isinstance(job, self.job_class):
                 job_instance = job
             else:
-                job_instance = Job.fetch(job_id, connection=connection)
+                job_instance = Job.fetch(job_id, connection=connection, serializer=self.serializer)
             job_instance.delete()
         return result
 
@@ -100,13 +103,37 @@ class BaseRegistry(object):
 
     def get_queue(self):
         """Returns Queue object associated with this registry."""
-        return Queue(self.name, connection=self.connection)
+        return Queue(self.name, connection=self.connection, serializer=self.serializer)
 
     def get_expiration_time(self, job):
         """Returns job's expiration time."""
         score = self.connection.zscore(self.key, job.id)
         return datetime.utcfromtimestamp(score)
 
+    def requeue(self, job_or_id):
+        """Requeues the job with the given job ID."""
+        if isinstance(job_or_id, self.job_class):
+            job = job_or_id
+            serializer = job.serializer
+        else:
+            serializer = self.serializer
+            job = self.job_class.fetch(job_or_id, connection=self.connection, serializer=serializer)
+
+        result = self.connection.zrem(self.key, job.id)
+        if not result:
+            raise InvalidJobOperation
+
+        with self.connection.pipeline() as pipeline:
+            queue = Queue(job.origin, connection=self.connection,
+                          job_class=self.job_class, serializer=serializer)
+            job.started_at = None
+            job.ended_at = None
+            job.exc_info = ''
+            job.save()
+            job = queue.enqueue_job(job, pipeline=pipeline)
+            pipeline.execute()
+        return job
+
 
 class StartedJobRegistry(BaseRegistry):
     """
@@ -130,20 +157,29 @@ class StartedJobRegistry(BaseRegistry):
         job_ids = self.get_expired_job_ids(score)
 
         if job_ids:
-            failed_job_registry = FailedJobRegistry(self.name, self.connection)
+            failed_job_registry = FailedJobRegistry(self.name, self.connection, serializer=self.serializer)
 
             with self.connection.pipeline() as pipeline:
                 for job_id in job_ids:
                     try:
                         job = self.job_class.fetch(job_id,
-                                                   connection=self.connection)
+                                                   connection=self.connection,
+                                                   serializer=self.serializer)
+                    except NoSuchJobError:
+                        continue
+
+                    retry = job.retries_left and job.retries_left > 0
+
+                    if retry:
+                        queue = self.get_queue()
+                        job.retry(queue, pipeline)
+
+                    else:
                         job.set_status(JobStatus.FAILED)
                         job.exc_info = "Moved to FailedJobRegistry at %s" % datetime.now()
                         job.save(pipeline=pipeline, include_meta=False)
                         job.cleanup(ttl=-1, pipeline=pipeline)
                         failed_job_registry.add(job, job.failure_ttl)
-                    except NoSuchJobError:
-                        pass
 
                 pipeline.zremrangebyscore(self.key, 0, score)
                 pipeline.execute()
@@ -207,27 +243,6 @@ class FailedJobRegistry(BaseRegistry):
         if not pipeline:
             p.execute()
 
-    def requeue(self, job_or_id):
-        """Requeues the job with the given job ID."""
-        if isinstance(job_or_id, self.job_class):
-            job = job_or_id
-        else:
-            job = self.job_class.fetch(job_or_id, connection=self.connection)
-
-        result = self.connection.zrem(self.key, job.id)
-        if not result:
-            raise InvalidJobOperation
-
-        with self.connection.pipeline() as pipeline:
-            queue = Queue(job.origin, connection=self.connection,
-                          job_class=self.job_class)
-            job.started_at = None
-            job.ended_at = None
-            job.save()
-            job = queue.enqueue_job(job, pipeline=pipeline)
-            pipeline.execute()
-        return job
-
 
 class DeferredJobRegistry(BaseRegistry):
     """
@@ -306,18 +321,34 @@ class ScheduledJobRegistry(BaseRegistry):
         return datetime.fromtimestamp(score, tz=timezone.utc)
 
 
+class CanceledJobRegistry(BaseRegistry):
+    key_template = 'rq:canceled:{0}'
+
+    def get_expired_job_ids(self, timestamp=None):
+        raise NotImplementedError
+
+    def cleanup(self):
+        """This method is only here to prevent errors because this method is
+        automatically called by `count()` and `get_job_ids()` methods
+        implemented in BaseRegistry."""
+        pass
+
+
 def clean_registries(queue):
     """Cleans StartedJobRegistry, FinishedJobRegistry and FailedJobRegistry of a queue."""
     registry = FinishedJobRegistry(name=queue.name,
                                    connection=queue.connection,
-                                   job_class=queue.job_class)
+                                   job_class=queue.job_class,
+                                   serializer=queue.serializer)
     registry.cleanup()
     registry = StartedJobRegistry(name=queue.name,
                                   connection=queue.connection,
-                                  job_class=queue.job_class)
+                                  job_class=queue.job_class,
+                                  serializer=queue.serializer)
     registry.cleanup()
 
     registry = FailedJobRegistry(name=queue.name,
                                  connection=queue.connection,
-                                 job_class=queue.job_class)
+                                 job_class=queue.job_class,
+                                 serializer=queue.serializer)
     registry.cleanup()
