@@ -4,36 +4,39 @@ import signal
 import time
 import traceback
 from datetime import datetime
+from enum import Enum
 from multiprocessing import Process
 
-from redis import Redis, SSLConnection
+from redis import SSLConnection, UnixDomainSocketConnection
 
 from .defaults import DEFAULT_LOGGING_DATE_FORMAT, DEFAULT_LOGGING_FORMAT
 from .job import Job
 from .logutils import setup_loghandlers
 from .queue import Queue
 from .registry import ScheduledJobRegistry
-from .utils import current_timestamp, enum
+from .serializers import resolve_serializer
+from .utils import current_timestamp
 
 SCHEDULER_KEY_TEMPLATE = 'rq:scheduler:%s'
 SCHEDULER_LOCKING_KEY_TEMPLATE = 'rq:scheduler-lock:%s'
 
 
-class RQScheduler(object):
+class SchedulerStatus(str, Enum):
+    STARTED = 'started'
+    WORKING = 'working'
+    STOPPED = 'stopped'
+
+
+class RQScheduler:
     # STARTED: scheduler has been started but sleeping
     # WORKING: scheduler is in the midst of scheduling jobs
     # STOPPED: scheduler is in stopped condition
 
-    Status = enum(
-        'SchedulerStatus',
-        STARTED='started',
-        WORKING='working',
-        STOPPED='stopped'
-    )
+    Status = SchedulerStatus
 
     def __init__(self, queues, connection, interval=1, logging_level=logging.INFO,
                  date_format=DEFAULT_LOGGING_DATE_FORMAT,
-                 log_format=DEFAULT_LOGGING_FORMAT):
+                 log_format=DEFAULT_LOGGING_FORMAT, serializer=None):
         self._queue_names = set(parse_names(queues))
         self._acquired_locks = set()
         self._scheduled_job_registries = []
@@ -48,6 +51,17 @@ class RQScheduler(object):
         connection_class = connection.connection_pool.connection_class
         if issubclass(connection_class, SSLConnection):
             self._connection_kwargs['ssl'] = True
+        if issubclass(connection_class, UnixDomainSocketConnection):
+            # The connection keyword arguments are obtained from
+            # `UnixDomainSocketConnection`, which expects `path`, but passed to
+            # `redis.client.Redis`, which expects `unix_socket_path`, renaming
+            # the key is necessary.
+            # `path` is not left in the dictionary as that keyword argument is
+            # not expected by `redis.client.Redis` and would raise an exception.
+            self._connection_kwargs['unix_socket_path'] = self._connection_kwargs.pop(
+                'path'
+            )
+        self.serializer = resolve_serializer(serializer)
 
         self._connection = None
         self.interval = interval
@@ -92,7 +106,7 @@ class RQScheduler(object):
         pid = os.getpid()
         self.log.info("Trying to acquire locks for %s", ", ".join(self._queue_names))
         for name in self._queue_names:
-            if self.connection.set(self.get_locking_key(name), pid, nx=True, ex=60):
+            if self.connection.set(self.get_locking_key(name), pid, nx=True, ex=self.interval + 60):
                 successful_locks.add(name)
 
         # Always reset _scheduled_job_registries when acquiring locks
@@ -115,7 +129,7 @@ class RQScheduler(object):
             queue_names = self._acquired_locks
         for name in queue_names:
             self._scheduled_job_registries.append(
-                ScheduledJobRegistry(name, connection=self.connection)
+                ScheduledJobRegistry(name, connection=self.connection, serializer=self.serializer)
             )
 
     @classmethod
@@ -140,10 +154,12 @@ class RQScheduler(object):
             if not job_ids:
                 continue
 
-            queue = Queue(registry.name, connection=self.connection)
+            queue = Queue(registry.name, connection=self.connection, serializer=self.serializer)
 
             with self.connection.pipeline() as pipeline:
-                jobs = Job.fetch_many(job_ids, connection=self.connection)
+                jobs = Job.fetch_many(
+                    job_ids, connection=self.connection, serializer=self.serializer
+                )
                 for job in jobs:
                     if job is not None:
                         queue.enqueue_job(job, pipeline=pipeline)
@@ -170,11 +186,11 @@ class RQScheduler(object):
             with self.connection.pipeline() as pipeline:
                 for name in self._queue_names:
                     key = self.get_locking_key(name)
-                    pipeline.expire(key, self.interval + 5)
+                    pipeline.expire(key, self.interval + 60)
                 pipeline.execute()
         else:
             key = self.get_locking_key(next(iter(self._queue_names)))
-            self.connection.expire(key, self.interval + 5)
+            self.connection.expire(key, self.interval + 60)
 
     def stop(self):
         self.log.info("Scheduler stopping, releasing locks for %s...",
